import torch
import torch.nn as nn
import pdb

def FMS_cost_function(gt, pred, weight_action_class=None, action_weight=1, deltaT_weight=1, location_weight=1, min_dict=None, max_dict=None, device="None",config=None):
    if config is None and config =="None":
        num_actions = 9
        num_deltaT = 101
        num_start_x = 101
        num_start_y = 101
    else:
        num_actions = config["num_actions"]
        num_deltaT = config["delta_T_bin"]
        num_start_x = config["start_x_bin"]
        num_start_y = config["start_y_bin"]
    action = gt[:,0].long()
    deltaT = gt[:,1].float()
    start_x = gt[:,2].float()
    start_y = gt[:,3].float()

    action_pred = pred[:,0:num_actions].float()
    deltaT_pred = pred[:,num_actions:num_actions+num_deltaT].float()
    start_x_pred = pred[:,num_actions+num_deltaT:num_actions+num_deltaT+num_start_x].float()
    start_y_pred = pred[:,num_actions+num_deltaT+num_start_x:].float()

    #apply log softmax to the prediction
    action_pred = nn.LogSoftmax(dim=1)(action_pred)
    deltaT_pred = nn.LogSoftmax(dim=1)(deltaT_pred)
    start_x_pred = nn.LogSoftmax(dim=1)(start_x_pred)
    start_y_pred = nn.LogSoftmax(dim=1)(start_y_pred)

    CEL_action_func = nn.CrossEntropyLoss(weight=weight_action_class,reduction ="none")
    CEL_action  = torch.mean(CEL_action_func(action_pred,action)) 
    CEL_deltaT_func = nn.CrossEntropyLoss(reduction ="none")
    CEL_deltaT = torch.mean(CEL_deltaT_func(deltaT_pred,(deltaT*100).long()))
    CEL_start_x_func = nn.CrossEntropyLoss(reduction ="none")
    CEL_start_x = torch.mean(CEL_start_x_func(start_x_pred,(start_x*100).long()))
    CEL_start_y_func = nn.CrossEntropyLoss(reduction ="none")
    CEL_start_y = torch.mean(CEL_start_y_func(start_y_pred,(start_y*100).long()))

    #unprocess the loss components
    if max_dict is not None and min_dict is not None:
        device = torch.device('cuda' if torch.cuda.is_available() else 'cpu') if device is None or device == 'None' else device
        deltaT_pred = torch.exp( torch.argmax(deltaT_pred,1)/100 * ( torch.tensor(max_dict["delta_T"], device=device) - torch.tensor(min_dict["delta_T"], device=device)) + torch.tensor(min_dict["delta_T"], device=device)) - torch.tensor(1e-6, device=device)
        deltaT = torch.exp( deltaT * ( torch.tensor(max_dict["delta_T"], device=device) - torch.tensor(min_dict["delta_T"], device=device)) + torch.tensor(min_dict["delta_T"], device=device)) - torch.tensor(1e-6, device=device)
        start_x_pred = torch.argmax(start_x_pred,1)/100  * ( torch.tensor(max_dict["start_x"], device=device) - torch.tensor(min_dict["start_x"], device=device)) + torch.tensor(min_dict["start_x"], device=device)
        start_y_pred = torch.argmax(start_y_pred,1)/100 * ( torch.tensor(max_dict["start_y"], device=device) - torch.tensor(min_dict["start_y"], device=device)) + torch.tensor(min_dict["start_y"], device=device)
        start_x = start_x * ( torch.tensor(max_dict["start_x"], device=device) - torch.tensor(min_dict["start_x"], device=device)) + torch.tensor(min_dict["start_x"], device=device)
        start_y = start_y * ( torch.tensor(max_dict["start_y"], device=device) - torch.tensor(min_dict["start_y"], device=device)) + torch.tensor(min_dict["start_y"], device=device)
        #ignore in backpropagation
        ACC_action = torch.mean((torch.argmax(action_pred,1)==action).float()).detach()
        AE_deltaT = torch.mean(torch.abs(deltaT_pred-deltaT)).detach()
        AE_start_x = torch.mean(torch.abs(start_x_pred-start_x)).detach()
        AE_start_y = torch.mean(torch.abs(start_y_pred-start_y)).detach()
    else:
        ACC_action = torch.tensor(-1, device=device)
        AE_deltaT = torch.tensor(-1, device=device)
        AE_start_x = torch.tensor(-1, device=device)
        AE_start_y = torch.tensor(-1, device=device)
    
    #f1 score
    pred_classes = torch.argmax(action_pred, dim=1)
    f1_action = calculate_f1_score(action, pred_classes, num_classes=num_actions)
        
    #cost function
    train_loss = action_weight*CEL_action + deltaT_weight*CEL_deltaT + location_weight*CEL_start_x + location_weight*CEL_start_y

    return train_loss,[CEL_action,CEL_deltaT,CEL_start_x,CEL_start_y,ACC_action,f1_action,AE_deltaT,AE_start_x,AE_start_y]

def calculate_f1_score(gt_classes, pred_classes, num_classes):
    f1_scores = []
    
    for i in range(num_classes):
        # True Positives (TP), False Positives (FP), False Negatives (FN)
        tp = torch.sum((pred_classes == i) & (gt_classes == i)).float()
        fp = torch.sum((pred_classes == i) & (gt_classes != i)).float()
        fn = torch.sum((pred_classes != i) & (gt_classes == i)).float()
        
        # Precision and Recall
        precision = tp / (tp + fp + 1e-8)  # Add small constant to avoid division by zero
        recall = tp / (tp + fn + 1e-8)     # Add small constant to avoid division by zero
        
        # F1 Score
        f1 = 2 * (precision * recall) / (precision + recall + 1e-8)  # Add small constant to avoid division by zero
        f1_scores.append(f1)
    
    # Average F1 Score across all classes
    avg_f1_score = torch.mean(torch.stack(f1_scores))
    
    return avg_f1_score


import pandas as pd
import numpy as np
import pdb

train_path= "/home/c_yeung/workspace6/python/openstarlab/Event/test/dataset/csv/train.csv"
train_df = pd.read_csv(train_path)

#count action=shot and success
# shot_df = train_df[(train_df["action"]=="shot") & (train_df["goal"]==1)]
# print("Number of successful shots: ", shot_df.shape[0])
# shot_df = train_df[(train_df["action"]=="shot") & (train_df["goal"]==0)]
# print("Number of failed shots: ", shot_df.shape[0])
'''
Number of successful shots:  581
Number of failed shots:  4995
'''

# train_df.action.value_counts()
'''
short_pass     168607
carry          147316
_               56481
high_pass       38831
cross            7841
dribble          6488
shot             5576
long_pass         517
period_over       225
game_over         225
'''


poss_len_data=[]
for i in train_df.match_id.unique():
    match_df = train_df[train_df["match_id"]==i]
    for j in match_df.poss_id.unique():
        poss_df = match_df[match_df["poss_id"]==j]
        poss_len_data.append(poss_df.shape[0])

#print the mean and standard deviation of the number of events in each possession
print("Mean number of events in each possession: ", np.mean(poss_len_data))
print("Standard deviation of the number of events in each possession: ", np.std(poss_len_data))
#print the maximum and minimum number of events in each possession
print("Maximum number of events in each possession: ", max(poss_len_data))
print("Minimum number of events in each possession: ", min(poss_len_data))
'''
Mean number of events in each possession:  7.65048423363609
Standard deviation of the number of events in each possession:  8.74643868143599
Maximum number of events in each possession:  127
Minimum number of events in each possession:  2

97.9% of possessions have 25.1434 events or less
99.9% of possessions have 33.8898 events or less
'''
pdb.set_trace()


import torch
import torch.nn as nn
import torch.nn.functional as F
from torch.utils.data import DataLoader
import math
import numpy as np
import pandas as pd
import json
import os
import time

from ..dataloaders.UEID_loader import UEID_loader
from ..trainers.FMS_train import FMS_train
from ..utils.UEID_preprocessing import UEID_preprocessing

import pdb

class BasicTransformer(nn.Transformer):
    #ref https://github.com/danielhocevar/Foundation-Model-for-Soccer/blob/main/models/transformer.py
    def __init__(self, input_dim, output_dim, num_actions=9, ninp=50, nhead=5, nhid=500, nlayers=4, dropout=0.2):
        """
        num_actions: dictionary length
        ninp: size of word embeddings
        nhead: number of heads in the encoder/decoder
        nhid: number of hidden units in hidden layers
        nlayers: number of hidden layers
        dropout: dropout probability
        """
        self.input_padding = nhead* (input_dim//nhead+1)-input_dim
        self.input_dim = input_dim + self.input_padding
        super(BasicTransformer, self).__init__(d_model=self.input_dim, nhead=nhead, dim_feedforward=nhid, num_encoder_layers=nlayers)
        self.model_type = 'Transformer'
        self.src_mask = None
        self.pos_encoder = PositionalEncoding(self.input_dim, dropout)

        self.input_emb = nn.Embedding(num_actions, ninp)
        self.ninp = ninp
        self.decoder = nn.Linear(self.input_dim, output_dim)

        self.init_weights()

    def _generate_square_subsequent_mask(self, sz):
        return torch.log(torch.tril(torch.ones(sz,sz)))

    def init_weights(self):
        initrange = 0.1
        nn.init.uniform_(self.input_emb.weight, -initrange, initrange)
        nn.init.zeros_(self.decoder.bias)
        nn.init.uniform_(self.decoder.weight, -initrange, initrange)

    def forward(self, src, has_mask=True):
        if has_mask:
            device = src.device
            if self.src_mask is None or self.src_mask.size(0) != len(src):
                mask = self._generate_square_subsequent_mask(len(src)).to(device)
                self.src_mask = mask
        else:
            self.src_mask = None

        src_action=src[:,:,0].long()
        src_action = self.input_emb(src_action) * math.sqrt(self.ninp)

        padding = torch.zeros(src.shape[0], src.shape[1], self.input_padding).to(src.device)

        src = torch.cat((src_action, src[:,:,1:], padding), dim=-1)
        src = self.pos_encoder(src)
        output = self.encoder(src, mask=self.src_mask)
        output = self.decoder(output[:,-1,:])
        return output

class PositionalEncoding(nn.Module):
    def __init__(self, d_model, dropout=0.1, max_len=5000):
        super(PositionalEncoding, self).__init__()
        self.dropout = nn.Dropout(p=dropout)

        pe = torch.zeros(max_len, d_model)
        position = torch.arange(0, max_len, dtype=torch.float).unsqueeze(1)
        div_term = torch.exp(torch.arange(0, d_model, 2).float() * (-math.log(10000.0) / d_model))
        pe[:, 0::2] = torch.sin(position * div_term)
        # Apply cos to odd indices in the array (1, 3, 5, ...)
        if d_model % 2 == 0:
            pe[:, 1::2] = torch.cos(position * div_term)
        else:
            pe[:, 1::2] = torch.cos(position * div_term[:(d_model // 2)])
        pe = pe.unsqueeze(0).transpose(0, 1)
        self.register_buffer('pe', pe)

    def forward(self, x):
        x = x + self.pe[:x.size(0), :]
        return self.dropout(x)

def train_FMS(current_time, config, train_df, valid_df, seq_len, features, epochs, batch_size,
            delta_T_bin, start_x_bin, start_y_bin,
            ninp, nhead, nhid, nlayers, dropout,
            num_actions=9, device=None, print_freq=10, patience=5, num_workers=4, optuna=False,
            learning_rate=0.01,eps=1e-16):

    # check if the train_df and valid_df are pandas dataframe
    if not isinstance(train_df, pd.DataFrame):
        #check if the train_df is a path
        if isinstance(train_df, str):
            train_df = pd.read_csv(train_df) if not config['test'] else pd.read_csv(train_df).head(1000)
        else:
            raise ValueError("train_df must be a pandas dataframe or a path to a csv file")
    if not isinstance(valid_df, pd.DataFrame):
        #check if the valid_df is a path
        if isinstance(valid_df, str):
            valid_df = pd.read_csv(valid_df) if not config['test'] else pd.read_csv(valid_df).head(1000)
        else:
            raise ValueError("valid_df must be a pandas dataframe or a path to a csv file")
        
    # Preprocess the data
    train_df, min_dict, max_dict = UEID_preprocessing(train_df)
    valid_df, _, _ = UEID_preprocessing(valid_df, min_dict, max_dict)

    # Define fixed hyperparameters 
    input_dim = len(features)-1+ ninp
    output_dim = num_actions + delta_T_bin + start_x_bin + start_y_bin

    #if not device is given, check if cuda is available
    device = torch.device('cuda' if torch.cuda.is_available() else 'cpu') if device is None or device == 'None' else device

    # Initialize model layers based on data
    model = BasicTransformer(input_dim, output_dim, num_actions, ninp, nhead, nhid, nlayers, dropout).to(device)
    
    # Create data loaders
    train_loader = UEID_loader(train_df, seq_len, features)
    valid_loader = UEID_loader(valid_df, seq_len, features)

    # Convert data loaders to torch DataLoader
    train_loader = DataLoader(train_loader, batch_size=batch_size, shuffle=True, drop_last=True, num_workers=num_workers)
    valid_loader = DataLoader(valid_loader, batch_size=batch_size, shuffle=False, drop_last=False, num_workers=num_workers)
    
    # Define optimizer
    optimizer = torch.optim.Adam(model.parameters(),lr=learning_rate,eps=eps)

    # Training logic here
    model_state_dict, best_epoch, train_losses, valid_losses, train_loss_components, valid_loss_components, flops = FMS_train(model, train_loader, valid_loader, min_dict, max_dict, optimizer, device, epochs, print_freq, patience,config=config)

    #create one csv file for the training and validation losses
    columns = ['train_loss', 'CEL_action', 'CEL_deltaT','CEL_start_x','CEL_start_y', 'ACC_action', 'F1_action', 'MAE_deltaT', 'MAE_x', 'MAE_y',
                'valid_loss', 'CEL_action_v', 'CEL_deltaT_v','CEL_start_x_v','CEL_start_y_v', 'ACC_action_v', 'F1_action_v', 'MAE_deltaT_v', 'MAE_x_v', 'MAE_y_v']
    data = [train_losses, np.array(train_loss_components)[:,0], np.array(train_loss_components)[:,1], np.array(train_loss_components)[:,2], np.array(train_loss_components)[:,3], np.array(train_loss_components)[:,4], np.array(train_loss_components)[:,5], np.array(train_loss_components)[:,6], np.array(train_loss_components)[:,7], np.array(train_loss_components)[:,8],
            valid_losses, np.array(valid_loss_components)[:,0], np.array(valid_loss_components)[:,1], np.array(valid_loss_components)[:,2], np.array(valid_loss_components)[:,3], np.array(valid_loss_components)[:,4], np.array(valid_loss_components)[:,5], np.array(valid_loss_components)[:,6], np.array(valid_loss_components)[:,7], np.array(valid_loss_components)[:,8]]
    data = np.array(data).T
    loss_df = pd.DataFrame(data, columns=columns)
    #round the loss_df to 4 decimal places
    loss_df = loss_df.round(4)

    
    #save the model and the loss_df
    if config['optuna']:
        method = "optuna"
    else:
        method = "train"
    i=1
    save_path = config['save_path']+f"/out/{method}/{current_time}/run_{i}/"
    while os.path.exists(save_path):
        i+=1
        save_path = config['save_path']+f"/out/{method}/{current_time}/run_{i}/"
    os.makedirs(save_path)

    model_save_path = save_path + f"_model_{best_epoch}.pth"
    loss_save_path = save_path + "_loss.csv"
    hyperparameters_save_path = save_path + "hyperparameters.json"
    model_stats_save_path = save_path + "_model_stats.txt"
    
    #save the min_dict and max_dict
    min_max_dict_path = config['save_path']+f"/out/{method}/{current_time}/min_max_dict.json"
    with open(min_max_dict_path, 'w') as f:
        json.dump({'min_dict':min_dict, 'max_dict':max_dict}, f, indent=4)

    #save all the hyperparameters values
    hyperparameters = {'current_time': current_time,
                        'train_path': config['train_path'], 'valid_path': config['valid_path'], 'save_path': config['save_path'],
                        'test': config['test'], 'batch_size': batch_size, 'num_epoch': config['num_epoch'],
                        'print_freq': print_freq, 'early_stop_patience': config['early_stop_patience'],
                        'dataloader_num_worker': config['dataloader_num_worker'], 'device': config['device'], 
                        'basic_features': config['basic_features'], 'other_features': config['other_features'],
                        'use_other_features': config['use_other_features'],
                        'features':features, 'num_actions': num_actions, 'seq_len': seq_len,
                        'optuna': config['optuna'], 'optuna_n_trials': config['optuna_n_trials'],
                        'learning_rate': learning_rate, 'eps': eps,
                        'delta_T_bin': delta_T_bin, 'start_x_bin': start_x_bin, 'start_y_bin': start_y_bin,
                        'ninp': ninp, 'nhead': nhead, 'nhid': nhid, 'nlayers': nlayers, 'dropout': dropout,
                        'input_dim': input_dim, 'output_dim': output_dim,
                        'model':model.__class__.__name__,
                        'best_epoch':best_epoch}
                
    torch.save(model_state_dict, model_save_path)
    loss_df.to_csv(loss_save_path, index=False)
    with open(hyperparameters_save_path, 'w') as f:
        json.dump(hyperparameters, f, indent=4)
    with open(model_stats_save_path, "w") as f:
        f.write(str(flops))
        
    print("Model and loss saved at", save_path)

    if optuna:
        return valid_losses[best_epoch]

def main(config):
    train_path = config['train_path']
    valid_path = config['valid_path']

    # Extract input dimensions from train_df
    if config['test']:
        seq_len = 10
        epochs = 2
        features = config['basic_features']+config['other_features']
    else:
        seq_len = config['seq_len']
        epochs = config['num_epoch']
        if config['use_other_features']:
            features = config['basic_features']+config['other_features']
        else:
            features = config['basic_features']
    
    batch_size = config['batch_size']
    num_actions = config['num_actions']
    print_freq = config['print_freq']
    early_stop_patience = config['early_stop_patience']
    dataloader_num_worker = config['dataloader_num_worker']
    learning_rate = config['learning_rate']
    eps = config['eps']
    eps = float(eps) if not isinstance(eps, float) else eps
    device = config['device']

    #tunable hyperparameters
    delta_T_bin = config['delta_T_bin']
    start_x_bin = config['start_x_bin']
    start_y_bin = config['start_y_bin']
    ninp = config['ninp']
    nhead = config['nhead']
    nhid = config['nhid']
    nlayers = config['nlayers']
    dropout = config['dropout']


    current_time = time.strftime("%Y%m%d_%H%M%S")
    train_FMS(current_time, config, train_path, valid_path, seq_len, features, epochs, batch_size,
                delta_T_bin, start_x_bin, start_y_bin,
                 ninp, nhead, nhid, nlayers, dropout,
                num_actions=num_actions,device=device,print_freq=print_freq, patience=early_stop_patience,num_workers=dataloader_num_worker,
                learning_rate=learning_rate,eps=eps)

def main_optuna(config):
    import optuna
    train_path = config['train_path']
    valid_path = config['valid_path']

    # Extract input dimensions from train_df
    if config['test']:
        seq_len = 10
        epochs = 2
        features = config['basic_features']+config['other_features']
    else:
        seq_len = config['seq_len']
        epochs = config['num_epoch']
        if config['use_other_features']:
            features = config['basic_features']+config['other_features']
        else:
            features = config['basic_features']
    
    num_actions = config['num_actions']
    print_freq = config['print_freq']
    early_stop_patience = config['early_stop_patience']
    dataloader_num_worker = config['dataloader_num_worker']
    device = config['device']

    delta_T_bin = config['delta_T_bin']
    start_x_bin = config['start_x_bin']
    start_y_bin = config['start_y_bin']

    def objective(trial):
        # tunable hyperparameters
        try:
            learning_rate=trial.suggest_float('learning_rate', float(config['learning_rate'][0]), float(config['learning_rate'][1])) if len(config['learning_rate'])>1 else config['learning_rate'][0]
            config['eps'] = [float(i) for i in config['eps']]
            eps=trial.suggest_float('eps',  float(config['eps'][1]), float(config['eps'][1])) if len(config['eps'])>1 else config['eps'][0]
            batch_size=trial.suggest_categorical('batch_size', config['batch_size'])  
          
            ninp=trial.suggest_categorical('ninp', config['ninp'])
            nhead=trial.suggest_categorical('nhead', config['nhead'])
            nhid=trial.suggest_categorical('nhid', config['nhid'])
            nlayers=trial.suggest_categorical('nlayers', config['nlayers'])
            dropout=trial.suggest_categorical('dropout', config['dropout'])
        except:
            print("Error in the yaml file")
            
        # Call your train function with these parameters
        valid_loss = train_FMS(current_time, config, train_path, valid_path, seq_len, features, epochs, batch_size,
                delta_T_bin, start_x_bin, start_y_bin,
                ninp, nhead, nhid, nlayers, dropout, optuna=True,
                num_actions=num_actions,device=device,print_freq=print_freq, patience=early_stop_patience,num_workers=dataloader_num_worker,
                learning_rate=learning_rate,eps=eps)
        
        return valid_loss


    n_trials = config['optuna_n_trials'] if not config['test'] else 2

    # Create an Optuna study object
    study = optuna.create_study(direction='minimize')
    
    # Optimize the objective function
    current_time = time.strftime("%Y%m%d_%H%M%S")
    study.optimize(objective, n_trials=n_trials)
    
    # Print the best trial results
    print("Best trial:")
    trial = study.best_trial
    best_trial_number = trial.number+1
    best_params = trial.params
    print(f"  Trial Number: {best_trial_number}")
    print(f"  Best Valid Loss: {trial.value:.4f}")
    print("  Params: ")
    for key, value in trial.params.items():
        print(f"    {key}: {value}")
    
    # Add the trial number to the parameters dictionary
    best_params_with_trial = {
        'best_trial_number': best_trial_number,
        'best_valid_loss': trial.value,
        **best_params
    }

    # Save the best hyperparameters with the trial number to a JSON file
    with open(config['save_path']+f'out/optuna/{current_time}/optuna_best_hyperparameters.json', 'w') as f:
        json.dump(best_params_with_trial, f, indent=4)

if __name__ == "__main__":
    import pandas as pd
    import argparse
    import yaml

    prase = argparse.ArgumentParser()
    prase.add_argument('--config_path', '-c', type=str, default=os.getcwd()+'/event/sports/soccer/models/model_yaml_test/train_FMS.yaml')
    args = prase.parse_args()
    config_path = args.config_path

    with open(config_path, 'r') as file:
        config = yaml.safe_load(file)

    if config['optuna']:
        main_optuna(config)
    else:
        main(config)

    '''
    all_features = ['match_id','poss_id','team','home_team','action','success','goal','home_score',
                'away_score','goal_diff','Period','Minute','Second','seconds','delta_T',
                'start_x','start_y','deltaX','deltaY','distance','dist2goal','angle2goal']
    basic_features = ['action', 'delta_T', 'start_x','start_y'] 
    derivable_features = ['team','home_team','action','success','seconds','delta_T','start_x','start_y','deltaX','deltaY','distance','dist2goal','angle2goal']
    other_features = ['team','home_team','success','seconds','deltaX','deltaY','distance','dist2goal','angle2goal']
    '''
    



import pandas as pd
import matplotlib.pyplot as plt
import numpy as np
import os

train_path= "/home/c_yeung/workspace6/python/openstarlab/Event/test/dataset/csv/train.csv"
train_df = pd.read_csv(train_path)

#plot the distribution of inter-event times
plt.figure()
plt.hist(train_df[train_df["delta_T"] != 0]["delta_T"], bins=100, density=True)
plt.title("Distribution of inter-event times")
plt.xlabel("Interevent time (s)")
plt.ylabel("Density")

#save the plot
save_path= "/home/c_yeung/workspace6/python/openstarlab/Event/test/dataset/plot"
os.makedirs(save_path, exist_ok=True)
plt.savefig(os.path.join(save_path, "interevent_time_dist.png"))

#calculate the mean and standard deviation of inter-event times
mean_interevent_time = np.mean(train_df[train_df["delta_T"] != 0]["delta_T"])
std_interevent_time = np.std(train_df[train_df["delta_T"] != 0]["delta_T"])
print("Mean inter-event time: ", mean_interevent_time)
print("Standard deviation of inter-event time: ", std_interevent_time)

#take log of inter-event times and plot the distribution
plt.figure()
plt.hist(np.log(train_df[train_df["delta_T"] != 0]["delta_T"]+1e-6), bins=100, density=True)
plt.title("Distribution of log inter-event times")
plt.xlabel("Log interevent time")
plt.ylabel("Density")

#save the plot
plt.savefig(os.path.join(save_path, "log_interevent_time_dist.png"))

#calculate the mean and standard deviation of log inter-event times
mean_log_interevent_time = np.mean(np.log(train_df[train_df["delta_T"] != 0]["delta_T"]+1e-6))
std_log_interevent_time = np.std(np.log(train_df[train_df["delta_T"] != 0]["delta_T"]+1e-6))
print("Mean log inter-event time: ", mean_log_interevent_time)
print("Standard deviation of log inter-event time: ", std_log_interevent_time)




import torch
from torch.utils.data import Dataset
import numpy as np
import pandas as pd
from typing import List
import pdb

class UEID_loader(Dataset):
    def __init__(self, dataset: pd.DataFrame, seq_len: int, features: List[int], target_features: List[str] = ['action', 'delta_T', 'start_x', 'start_y']):
        self.df = dataset.reset_index(drop=True)
        self.seq_len = seq_len
        self.features = features
        self.target_features = target_features
        
        # Validate sequence length and indices
        self.valid_indices = []
        for i in range(len(self.df) - self.seq_len):
            if (self.df.loc[i, "match_id"] == self.df.loc[i + self.seq_len, "match_id"] and 
                self.df.loc[i, "Period"] == self.df.loc[i + self.seq_len, "Period"]):
                self.valid_indices.append(i)
        
    def __len__(self) -> int:
        return len(self.valid_indices)
    
    def __getitem__(self, idx: int):
        start_idx = self.valid_indices[idx]
        end_idx = start_idx + self.seq_len
        # Ensure that indices are within bounds
        if end_idx >= len(self.df):
            raise IndexError("Sequence end index exceeds the length of the dataframe")
        
        # Get the input sequence and target sequence
        input_seq = self.df.loc[start_idx:end_idx-1, self.features].values
        target_seq = self.df.loc[end_idx, self.target_features].values
        
        #convert the value to float
        input_seq = input_seq.astype(float)
        target_seq = target_seq.astype(float)

        # Convert to torch tensors
        input_seq = torch.tensor(input_seq, dtype=torch.float32)
        target_seq = torch.tensor(target_seq, dtype=torch.float32)

        return input_seq, target_seq, end_idx

class UEID_Simulation_loader(Dataset):
    def __init__(self, dataset: pd.DataFrame, seq_len: int, features: List[int], target_features: List[str] = ['action', 'delta_T', 'start_x', 'start_y']):
        self.df = dataset.reset_index(drop=True)
        self.seq_len = seq_len
        self.features = features
        self.target_features = target_features
        
        #fiter out possesion without end action
        valid_possessions = self.df.groupby(['match_id', 'poss_id']).filter(lambda x: x["action"].iloc[-1] == 3) # 3 is the end action
        self.df = valid_possessions
        #print numnber of row dropped
        # print(f"Number of rows dropped: {len(dataset) - len(self.df)}")

        # Validate sequence length and indices
        self.valid_indices = []
        for i in range(len(self.df) - self.seq_len):
            if (self.df.loc[i, "match_id"] == self.df.loc[i + self.seq_len, "match_id"] and 
                self.df.loc[i, "Period"] == self.df.loc[i + self.seq_len, "Period"]):
                self.valid_indices.append(i)
        
    def __len__(self) -> int:
        return len(self.valid_indices)
    
    def __getitem__(self, idx: int):
        start_idx = self.valid_indices[idx]
        end_idx = start_idx + self.seq_len
        # Ensure that indices are within bounds
        if end_idx >= len(self.df):
            raise IndexError("Sequence end index exceeds the length of the dataframe")
        
        # Get the input sequence and target sequence
        input_seq = self.df.loc[start_idx:end_idx-1, self.features].values
        
        # Convert to torch tensors
        input_seq = torch.tensor(input_seq, dtype=torch.float32)

        return input_seq, end_idx


import pandas as pd
import numpy as np
import pdb

def UEID_preprocessing(df, min_dict_input=None, max_dict_input=None):
    # Drop the game_over and period_over rows:
    df = df[df["action"] != "game_over"]
    df = df[df["action"] != "period_over"]

    # Mapping for team and action columns
    team_dict = {'Granada': 0, 'Atlético Madrid': 1, 'Celta Vigo': 2, 'Osasuna': 3, 
                 'Valencia': 4, 'Sevilla': 5, 'Mallorca': 6, 'Las Palmas': 7, 'Cádiz': 8, 
                 'Deportivo Alavés': 9, 'Real Sociedad': 10, 'Girona': 11, 'Villarreal': 12, 
                 'Real Betis': 13, 'Barcelona': 14, 'Getafe': 15, 'Rayo Vallecano': 16, 
                 'Almería': 17, 'Athletic Club': 18, 'Real Madrid': 19}
    action_dict = {'short_pass': 0, 'carry': 1, 'high_pass': 2, '_': 3, 'cross': 4, 
                   'long_pass': 5, 'shot': 6, 'dribble': 7}

    # Map the team and action columns to integers
    df["team_map"] = df["team"].copy().map(team_dict) 
    if df["team_map"].isnull().sum() == 0:
        df["team"] = df["team_map"]
    df.drop(columns=["team_map"], inplace=True)
    df["action"] = df["action"].map(action_dict)

    #if action is 6 and success is 1, then change action to 8
    df.loc[(df["action"] == 6) & (df["goal"] == 1), "action"] = 8

    # Initialize min and max dictionaries
    if min_dict_input is None or max_dict_input is None:
        min_dict = {}
        max_dict = {}
        
        # Calculate minimums and maximums for min-max normalization
        features_to_normalize = ["seconds", "start_x", "start_y", "deltaX", "deltaY", "distance", "dist2goal", "angle2goal",'home_score','away_score']
        for feature in features_to_normalize:
            min_dict[feature] = float(df[feature].min())
            max_dict[feature] = float(df[feature].max())
    else:
        min_dict = min_dict_input
        max_dict = max_dict_input

    

    # Apply min-max normalization
    for feature in ["seconds", "start_x", "start_y", "deltaX", "deltaY", "distance", "dist2goal", "angle2goal",'home_score','away_score']:
        if max_dict[feature] - min_dict[feature] != 0:
            df[feature] = (df[feature] - min_dict[feature]) / (max_dict[feature] - min_dict[feature])
        else:
            df[feature] = 0
        #ensure the values are between 0 and 1
        df[feature] = df[feature].apply(lambda x: 0 if x < 0 else 1 if x > 1 else x)


    # Apply logarithmic transformation and min-max normalization to delta_T
    df["delta_T"] = df["delta_T"].apply(lambda x: np.log(x + 1e-6))
    if min_dict_input is None and max_dict_input is None:
        min_dict["delta_T"] = df["delta_T"].min()
        max_dict["delta_T"] = df["delta_T"].max()
    else:
        min_dict["delta_T"] = min_dict_input["delta_T"]
        max_dict["delta_T"] = max_dict_input["delta_T"]
        
    df["delta_T"] = (df["delta_T"] - min_dict["delta_T"]) / (max_dict["delta_T"] - min_dict["delta_T"])
    #ensure the values are between 0 and 1
    df["delta_T"] = df["delta_T"].apply(lambda x: 0 if x < 0 else 1 if x > 1 else x)
    
    return df, min_dict, max_dict




# from /home/c_yeung/workspace6/python/openstarlab/Event/event/sports/soccer/main_class_soccer/main.py
from .soccer.main_class_soccer.main import event_model_soccer

class Event_Model:
    soccer_event_model = ['FMS','LEM_action','LEM','MAJ','NMSTPP','Seq2Event']
    other_model = []

    def __new__(cls, event_model, *args, **kwargs):
        if event_model in cls.soccer_event_model:
            return event_model_soccer(event_model, *args, **kwargs)
        elif event_model in cls.other_model:
            raise NotImplementedError('other model not implemented yet')
        else:
            raise ValueError(f'Unknown event model: {event_model}')


if __name__ == '__main__':
    import os
    #Test FMS
    # model = 'Seq2Event'
    # model = Event_Model(model, f'/home/c_yeung/workspace6/python/openstarlab/Event_Pretrain/model_yaml_test/train_{model}_optuna.yaml')
    # model.train()
    #Example only, run the inference function after training
    # model_path = os.getcwd()+'/test/model/FMS/out/train/20240922-181450/run_1/_model_1.pth'
    # model_config = os.getcwd()+'/test/model/FMS/out/train/20240922-181450/run_1/hyperparameters.json'
    # model.inference(model_path, model_config) #simple inference
    # model.inference(model_path, model_config, simulation=True, random_selection=True, max_iter=20) #simulation with evaluation
    # model = Event_Model('FMS', os.getcwd()+'/event/sports/soccer/models/train_FMS_optuna.yaml')
    # model.train()
    print('Done')


import pandas as pd
import os
import pdb

def split_dataset(id_matching_csv:str,event_data_csv:str):
    #read id_list
    id_list = pd.read_csv(id_matching_csv)

    #order the id_list by the match_date_statsbomb
    id_list = id_list.sort_values("match_date_statsbomb").reset_index(drop=True)

    #split the id_list into train, validation and test with 60%, 20%, 20% respectively
    train_size = int(0.6 * len(id_list)) #228
    val_size = int(0.2 * len(id_list)) #76
    test_size = len(id_list) - train_size - val_size #76
    print(f"train_size: {train_size}, val_size: {val_size}, test_size: {test_size}")

    train_id_list = id_list.iloc[:train_size].match_id_statsbomb.to_list()
    val_id_list = id_list.iloc[train_size:train_size+val_size].match_id_statsbomb.to_list()
    test_id_list = id_list.iloc[train_size+val_size:].match_id_statsbomb.to_list()

    #load the preprocessed event data
    event_data = pd.read_csv(event_data_csv)

    #split the event_data into train, validation and test
    train_event_data = event_data[event_data.match_id.isin(train_id_list)]
    val_event_data = event_data[event_data.match_id.isin(val_id_list)]
    test_event_data = event_data[event_data.match_id.isin(test_id_list)]

    return train_event_data, val_event_data, test_event_data

if __name__ == "__main__":
    id_matching_csv = "/home/c_yeung/workspace6/python/openstarlab/Event/test/dataset/csv/id_matching.csv"
    event_data_csv = "/home/c_yeung/workspace6/python/openstarlab/Event/test/dataset/csv/data.csv"
    save_dir = "/home/c_yeung/workspace6/python/openstarlab/Event/test/dataset/csv"
    train_event_data, val_event_data, test_event_data = split_dataset(id_matching_csv, event_data_csv)
    train_event_data.to_csv(os.path.join(save_dir, "train.csv"), index=False)
    val_event_data.to_csv(os.path.join(save_dir, "valid.csv"), index=False)
    test_event_data.to_csv(os.path.join(save_dir, "test.csv"), index=False)



import pandas as pd
import pdb

df_path= "/home/c_yeung/workspace6/python/openstarlab/Event/test/dataset/csv/train.csv"
df = pd.read_csv(df_path)

# found the len of the possession with "poss_id"
possession_len = df.groupby(['match_id', 'poss_id']).size().reset_index(name='len')
#get the stats of the possession length
possession_len_stats = possession_len['len'].describe()
#get the 99% percentile of the possession length
possession_len_99 = possession_len['len'].quantile(0.99)
pdb.set_trace() 
"""
for train.csv, possession_len_stats is:
count    56481.000000
mean         7.650484
std          8.746516
min          2.000000
25%          2.000000
50%          4.000000
75%          9.000000
max        127.000000
"""